from pandas import DataFrame, DateOffset, date_range, read_csv, Series
from pandas import Timedelta, Timestamp, to_datetime


class DataWrapper:
    def __init__(self):
        self.FIRST_DATETIME = '2011-01-01 01'
        self.LAST_DATETIME = '2018-01-01 00'
        self.TIME_ZONE = 'Europe/Oslo'
        
        self.WEIGHTS = {'namsos': 0.19, 'steinkjer': 0.5, 'stjordal': 0.31}
        
        self.COL_NAMES = ['year',
                     'month',
                     'day',
                     'hour',
                     'date_string',
                     'power',
                     'power_other',
                     'Kalman',
                     'namsos_temp',
                     'steinkjer_temp',
                     'stjordal_temp',
                     'namsos_temp_pred',
                     'steinkjer_temp_pred',
                     'stjordal_temp_pred',
                     'namsos_wind_speed',
                     'stjordal_wind_speed',
                     'steinkjer_wind_speed',
                     'namsos_wind_speed_pred',
                     'steinkjer_wind_speed_pred',
                     'stjordal_wind_speed_pred',
                     'namsos_radiation',
                     'steinkjer_radiation',
                     'stjordal_radiation',
                     'namsos_radiation_pred',
                     'steinkjer_radiation_pred',
                     'stjordal_radiation_pred',
                     ]
        
        _lower_time_limit = to_datetime('2011-01-01 01').tz_localize(self.TIME_ZONE)
        _upper_time_limit = to_datetime(self.LAST_DATETIME).tz_localize(self.TIME_ZONE)
        
        self.start_date = Timestamp(_lower_time_limit)
        self.end_date = Timestamp(_upper_time_limit)

        time_index = date_range(start=self.FIRST_DATETIME, end=self.LAST_DATETIME,
                                freq='H', tz=self.TIME_ZONE)

        self._file_loc = 'C:\\Users\\Torfinn\\Documents\\masteroppgave\\data.csv'

        self.dataset = read_csv(self._file_loc,
                                   sep=',',
                                   skiprows=2,
                                   names=self.COL_NAMES,
                                   index_col=False,
                                   )
        
        self.dataset.index = time_index


    def get(self, requests, start=None, drop_na=True,
            hours=0, days=0, weeks=0, months=0, years=0,
            **deprecated_kwargs):

        start_datetime = self.start_date if start is None \
            else Timestamp(start)

        duration_dict = {'hours': hours, 'days': days, 'weeks': weeks,
                         'months': months, 'years': years}

        if sum(duration_dict.values()) == 0:
            end_datetime = self.end_date
        else:
            duration = DateOffset(**duration_dict)
            end_datetime = start_datetime + duration - DateOffset(hours=1)

        interval = {'start': start_datetime, 'end': end_datetime}

        request_list = requests if type(requests) is list else [requests]

        requested_data = {}
        for request in request_list:
            series = self._get_single_data_series(request, interval, self.WEIGHTS)

            if ' shift_' in request:
                series = self._shift_data_in_time(series, request)

            requested_data[request] = series

        df = DataFrame(requested_data)

        if drop_na:
            df = df.dropna()

        return df


    def shift(self, df: DataFrame, by_hours: int, **get_method_kwargs):
        offset = DateOffset(hours=by_hours)
        new_start = df.index.min() - offset
        duration = len(df) + by_hours

        request = df.columns.ravel().tolist()
        bigger_df = self.get(request, start=new_start,
                             hours=duration, **get_method_kwargs)

        shifted = bigger_df.shift(by_hours, freq='H').dropna()
        return shifted.loc[df.index]


    def _get_single_data_series(self, request, interval: dict, weights: dict):
        loc = request.find(' ')
        data_request = request if loc == -1 else request[:loc]

        weighted_data = ['temp', 'temp_pred', 'wind_speed',
                         'wind_speed_pred', 'radiation', 'radiation_pred']

        time_data = self.dataset.index._datetimelike_ops

        if any(data_request == s for s in self.dataset.columns):
            return self._get_standard_data(data_request, interval)

        elif any(data_request == s for s in weighted_data):
            return self._get_weighted_data(data_request, interval, weights)

        elif data_request in time_data:
            return self._get_datetimelike_info(data_request, interval)

        else:
            raise ValueError('The data type "{}" is not defined.'
                             .format(str(request)))


    def _get_standard_data(self, request: str, interval: dict):
        return self.dataset[interval['start']:interval['end']].eval(request)


    def _get_weighted_data(self, request: str, interval: dict, weights: dict):
        df = self.dataset[interval['start']:interval['end']]

        data_namsos = df.eval('namsos_' + request) * weights['namsos']
        data_steinkjer = df.eval('steinkjer_' + request) * weights['steinkjer']
        data_stjordal = df.eval('stjordal_' + request) * weights['stjordal']

        return data_namsos + data_steinkjer + data_stjordal


    def _get_datetimelike_info(self, request, interval):
        indices = self.dataset[interval['start']:interval['end']].index
        return Series(getattr(indices, request), index=indices)


    def _get_all_data_of_type(self, request: str, weights: dict):
        complete_span = {'start': self.start_date, 'end': self.end_date}
        return self._get_single_data_series(request, complete_span, weights)


    def _shift_data_in_time(self, series, request):
        period_substring = self._extract_subsequent_substring(request, ' shift_')

        period_type, period_amount = period_substring.split('=')
        delta_kwarg = {period_type: int(period_amount)}
        delta = Timedelta(**delta_kwarg)

        hours = int(delta / Timedelta(hours=1))

        return series.shift(hours)
    
    
    @staticmethod
    def _extract_subsequent_substring(larger_string: str, query_string: str):
        start_loc = larger_string.find(query_string) + len(query_string)
        end_loc = larger_string.find(' ', start_loc)

        if end_loc != -1:
            return larger_string[start_loc: end_loc]
        return larger_string[start_loc:]