from datetime import datetime, timedelta
import matplotlib
from matplotlib import pyplot as plt
import matplotlib.dates as mdates
import pandas
from pandas import date_range

from Library.error_functions import MAE, MAPE
from Library.helper_functions import generate_holidays

matplotlib.rcParams.update({'font.size': 14})


def errors_by_hour_plot(power, predictions):
    l1 = plot_error_by_hour(power, predictions['Ensemble'], new_figure=False,
                            label='Mean of three best models', color='C0')
    l2 = plot_error_by_hour(power, predictions['MIMO MLP'], new_figure=False,
                            label='MIMO MLP', color='C1')
    l3 = plot_error_by_hour(power, predictions['SMSO MLP'], new_figure=False,
                            label='SMSO MLP', color='C2')
    l4 = plot_error_by_hour(power, predictions['LSTM Encoder-decoder'], new_figure=False,
                            label='LSTM Encoder-Decoder', color='C3')
    l5 = plot_error_by_hour(power, predictions['Direct SVR'], new_figure=False,
                            label='Direct SVR', color='C4')
    l6 = plot_error_by_hour(power, predictions['NTE Kalman'], new_figure=False,
                            label='Kalman filter (NTE)', color='C5')
    l7 = plot_error_by_hour(power, predictions['Direct Kalman'], new_figure=False,
                            label='Direct Kalman filter', color='C6')
    
    legend1 = plt.legend(handles=[l1, l2], loc='lower left', prop={'size': 10})
    ax = plt.gca().add_artist(legend1)
    plt.legend(handles=[l3, l4, l5, l6, l7], loc='lower right', prop={'size': 10})
        
    plt.tight_layout()
    plt.ylim([3, 12])
    plt.subplot_tool()
    
    
def generate_diagnostics(power, predictions):
    print(print_sorted_daily_errors(power, predictions))
    print('\n')
    print('MAPE:', round(MAPE(power.values, predictions.values), 2), '%')
    print('\n')
    print('MAE:', round(MAE(power.values, predictions.values), 2))
    print('\n')
    print_error_by_month(power, predictions)
    print('\n')
    print_error_by_weekday(power, predictions)
    print('\n')
    plot_error_by_hour(power, predictions)
    print('\n')
    # plot_error_by_time_of_week(power, predictions)
    
    residuals = power - predictions
    fig = plt.figure(figsize=(9, 4.2))
    ax = fig.add_subplot(1,1,1)  
    plt.plot(residuals)
    plt.xlabel('Time of year')
    plt.ylabel('Residual [MW]')
    plt.tight_layout()
    
    ax.xaxis.set_major_locator(mdates.MonthLocator(interval=1))
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%m'))
    
    plt.subplot_tool()
    
    
def history_plot(history, ymax=0.06):
    plt.figure()
    plt.plot(history.history['loss'], label='Training set')
    plt.plot(history.history['val_loss'], label='Test set')
    plt.xlabel("Epoch")
    plt.ylabel("Error (MSE)")
    plt.ylim([0, ymax])
    plt.legend(loc='upper right')
    
    
def plot_comparison(power, predictions, start_date=datetime(2017, 1, 1, 1),
                    end_date=datetime(2017, 12, 30, 0)):
    plt.figure(figsize=(9, 3.5))
    
    dates = date_range(start_date, end_date, freq='D')
    
    ax = power[start_date:end_date].tz_localize(None).plot()
    ax.set_xticks(dates)
    ax.set_ylabel("Power [MW]")
    
    ax = predictions[start_date:end_date].tz_localize(None).plot()
    ax.set_xticks(dates)
    
    plt.subplot_tool()
    
    print('MAPE:', round(MAPE(power[start_date:end_date].values, predictions[start_date:end_date].values),2), '%')
    print('MAE:', round(MAE(power[start_date:end_date].values, predictions[start_date:end_date].values),2))
    
    plt.tight_layout()

    #plt.legend(['Actual', 'Predictions'], bbox_to_anchor=(0.49, 0.65))
    plt.legend(['Actual', 'Predictions'], bbox_to_anchor=(0.025, 0.7))
    #plt.ylim([160, 360])
    # plt.legend(['Actual', 'Predictions'], loc='top left')


def print_error_by_month(power, predictions):
    MONTHS_IN_YEAR = 12
    
    months = ['Jan', 'Feb', 'March', 'April', 'May', 'June',
              'July', 'August', 'Sept', 'Oct', 'Nov', 'Dec']
    
    for month in range(1, MONTHS_IN_YEAR+1):
        power_this_month = power[power.index.month == month]
        predictions_this_month = predictions[predictions.index.month == month]
        print(months[month-1], '\t', round(MAE(power_this_month.values, predictions_this_month.values),2))
    
    
def plot_error_by_hour(power, predictions, new_figure=True, label=None, color=None):
    HOURS_IN_DAY = 24
    
    errors_by_hour = []
    
    for hour in range(1, HOURS_IN_DAY):
        power_this_hour = power[power.index.hour == hour]
        predictions_this_hour = predictions[predictions.index.hour == hour]
        
        errors_by_hour.append(round(MAE(power_this_hour.values, predictions_this_hour),2))
        
    power_this_hour = power[power.index.hour == 0]
    predictions_this_hour = predictions[predictions.index.hour == 0]
        
    errors_by_hour.append(round(MAE(power_this_hour.values, predictions_this_hour),2))
            
    if new_figure:
        plt.figure()
        
    l, = plt.plot(range(1,25), errors_by_hour, label=label, color=color)
    plt.xlabel('Hour of day')
    plt.ylabel('MAE [MW]')
    
    return l
    
    
def print_error_by_weekday(power, predictions):
    DAYS_IN_WEEK = 7
    
    weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    
    for weekday in range(0, DAYS_IN_WEEK):
        power_this_month = power[power.index.weekday == weekday]
        predictions_this_month = predictions[predictions.index.weekday == weekday]
        print(weekdays[weekday], '\t', round(MAE(power_this_month.values, predictions_this_month.values),2))
        
    start_date = predictions.first_valid_index().date()
    end_date = predictions.last_valid_index().date()
    
    holidays = generate_holidays(start_date, end_date)
    
    is_holiday = (predictions.index.date == holidays[0])
    for i in range(1, len(holidays)):
        is_holiday += (predictions.index.date == holidays[i])
        
    power = power[is_holiday]
    predictions = predictions[is_holiday]
    
    print('Holidays', round(MAE(power, predictions),2))
    
    
def print_errors_by_weekday(power, predictions):
    print('\nDirect FFNN')
    print_error_by_weekday(power, predictions['Direct MLP'])
    print('\nDirect SVM')
    print_error_by_weekday(power, predictions['Direct SVR'])
    print('\nMIMO FFNN')
    print_error_by_weekday(power, predictions['MIMO MLP'])
    print('\nLSTM RNN (Encoder-Decoder)')
    print_error_by_weekday(power, predictions['LSTM Encoder-decoder'])
    print('\nKalman filter')
    print_error_by_weekday(power, predictions['Kalman'])
    print('\nEnsemble')
    print_error_by_weekday(power, predictions['Ensemble'])
    
    
def print_sorted_daily_errors(power, predictions):
    start_date = power.index[0].date()
    end_date = power.index[-2].date()
    
    dates = pandas.date_range(start_date, end_date, tz='Europe/Oslo')
    
    errors_by_day = pandas.DataFrame(index=dates)
    errors_by_day['MAPE'] = [0.0]*len(dates)
    
    i = 0
    for day in errors_by_day.index:
        errors_by_day['MAPE'][i] = round(MAPE(power[day+timedelta(hours=1):day+timedelta(days=1)].values,
                   predictions[day+timedelta(hours=1):day+timedelta(days=1)].values),2)
        i += 1
    
    errors_by_day = errors_by_day.sort_values(by='MAPE', ascending=False)
    print(errors_by_day)