from datetime import date, datetime, timedelta
from keras.models import Sequential
from keras.layers import Dense
from keras.optimizers import Adam, SGD
from numpy import identity, zeros
from numpy import transpose as t
from numpy.linalg import inv
from pandas import concat, DataFrame, DatetimeIndex, Series
import pytz
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVR
from time import time

from Library.error_functions import MAE, MAPE
from Library.diagnostic_functions import generate_diagnostics, history_plot
from Library.helper_functions import generate_holidays, write_to_file

from Models.model import Model

HOURS_PER_DAY = 24
DAYS_PER_WEEK = 7
HOURS_PER_WEEK = HOURS_PER_DAY * DAYS_PER_WEEK


class Direct_model(Model):
    def __init__(self, dw, model_type):
        self.dw = dw
        self.model_type = model_type
        
    
    def fit(self, test_year, validation='Expanding',
            training_start=None, training_end=None, test_start=None, test_end=None):
        self.current_hour = 10
        self.power_lag = 5
        self.min_hours_ahead = 15
        self.max_hours_ahead = 38
        
        self.training_start = training_start
        self.training_end = training_end
        self.test_start = test_start
        self.test_end = test_end
        
        self.test_year = test_year
        
        self.validation = validation
        
        training_start_time = time()
        
        if self.model_type == 'Kalman':
            self.H_scalers = []
            self.y_scalers = []
        if self.model_type == 'MLP':
            self.histories = []
        self.models = []
        for hours_ahead in range(self.min_hours_ahead, self.max_hours_ahead+1):
            print('Hours ahead: ', hours_ahead)
            
            self._load_X(hours_ahead)
            self._load_y(hours_ahead)
        
            if 'Manual' in validation:
                self._split_data_manually(training_start, training_end, 
                                          test_start, test_end)
            else:
                self._split_data(hours_ahead)
                
            self._scale_data()
            
            if self.model_type == 'Lin_reg':
                model = self._fit_Lin_reg()
            elif self.model_type == 'MLP':
                model = self._fit_MLP()
            elif self.model_type == 'SVR':
                model = self._fit_SVR()
            elif self.model_type == 'Kalman':
                model = self._fit_Kalman(hours_ahead)
            
            self.models.append(model)
            
        training_end_time = time()
        self.training_time = training_end_time - training_start_time
        
        
    def _fit_Kalman(self, hours_ahead):
        a = self.X_training[self.X_training.index.year == (self.test_year - 1)]
        b = self.y_training[self.y_training.index.year == (self.test_year - 1)]
        
        H = self.X_training[self.X_training.index.year != (self.test_year - 1)]
        Y = self.y_training[self.y_training.index.year != (self.test_year - 1)]
        
        H = H.append(a)
        Y = Y.append(b)
        
        self.H = H
        self.Y = Y
        
        H_scaler = StandardScaler()
        H_scaler.fit_transform(H.values)
        self.H_scalers.append(H_scaler)
        H = DataFrame(H_scaler.transform(H.values), index=H.index)
        
        y_scaler = StandardScaler()
        y_scaler.fit_transform(Y.values)
        self.y_scalers.append(y_scaler)
        Y_scaled = DataFrame(y_scaler.transform(Y.values), index=Y.index)
        
        N = H.shape[1]
        
        omega = 1
        
        S = zeros(N)
        V = identity(N)
        
        for i in range(0, len(H)):
            timepoint = H.index[i]
            
            h = H[timepoint:timepoint].values
            y = Y_scaled[timepoint:timepoint][0][0]
            
            # Filter
            K = (V @ t(h)) @ inv(h @ V @ t(h) + omega)
            S_prev = S
            S = S_prev + K @ (y - h @ S_prev)
            V = (identity(N) - K @ h) @ V
            
        return [S_prev, S, V]
    
        
    def _fit_Lin_reg(self):
        model = LinearRegression()
        model.fit(self.X_training_scaled.values,
                          self.y_training_scaled.values.ravel())
        
        return model
        
        
    def _fit_MLP(self):
        model = Sequential()
        model.add(Dense(40, input_dim=self.X.shape[1], kernel_initializer='normal', activation='relu'))
        model.add(Dense(1, kernel_initializer='normal'))
        model.compile(loss='mean_squared_error', optimizer=Adam(lr=0.0001))
        
        history = model.fit(self.X_training_scaled.values,
                            self.y_training_scaled.values, epochs=2000,
                            validation_data=(self.X_test_scaled.values,
                                             self.y_test_scaled.values),
                            batch_size=256, verbose=0, shuffle=True)
        self.histories.append(history)
        
        return model
        
        
    def _fit_SVR(self):
        model = SVR(C=3, epsilon=0.1, gamma=0.01)
        model.fit(self.X_training_scaled.values,
                          self.y_training_scaled.values.ravel())
        
        return model
    
        
    def forecast(self):
        self.rolling_forecast = False
        
        update_start_time = time()  
        
        predictions = DataFrame()
        
        for hours_ahead in range(self.min_hours_ahead, self.max_hours_ahead+1):
            self._load_X(hours_ahead)
            self._load_y(hours_ahead)
            
            if 'Manual' in self.validation:
                self._split_data_manually(self.training_start, self.training_end, 
                                          self.test_start, self.test_end)
            else:
                self._split_data(hours_ahead)
                
            self._scale_data()
            
            model = self.models[hours_ahead-self.min_hours_ahead]
            
            y_hat_scaled = model.predict(self.X_test_scaled.values)
            y_hat = DataFrame(self.y_scaler.inverse_transform(y_hat_scaled))
            y_hat.index = self.y_test.index
            
            predictions = concat([predictions, y_hat])
            
        self.y_hat = predictions.sort_index()[0]
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
        self.update_time = update_time / len(self.y_test)
        
        return self.y_hat
    
    
    def rolling_forecast(self):
        update_start_time = time()
        
        self.y_hat = Series()
        for hours_ahead in range(self.min_hours_ahead, self.max_hours_ahead+1):   
            print(hours_ahead)
            self._load_X(hours_ahead)
            self._load_y(hours_ahead)
            self._split_data(hours_ahead)
            self._scale_data()
            
            model = self.models[hours_ahead-self.min_hours_ahead]
            
            if self.model_type == 'Kalman':
                self._rolling_forecast_Kalman(model, hours_ahead)
            elif self.model_type == 'MLP':
                self._rolling_forecast_MLP(model)
            
        self.y_hat = self.y_hat.sort_index()
                
        update_end_time = time()
        update_time = update_end_time - update_start_time
        self.update_time = update_time / len(self.y_test)
        
        return self.y_hat
    
    
    def _rolling_forecast_Kalman(self, model, hours_ahead):
        start_date = datetime(self.test_year,1,1,1)
        end_date = datetime(self.test_year+1,1,1,0)
        
        H = self.X[start_date:end_date]
        Y = self.y[start_date:end_date]
        
        H_scaler = StandardScaler()
        H_scaler.fit_transform(H.values)
        self.H_scalers.append(H_scaler)
        H = DataFrame(H_scaler.transform(H.values), index=H.index)
        
        y_scaler = StandardScaler()
        y_scaler.fit_transform(Y.values)
        self.y_scalers.append(y_scaler)
        Y_scaled = DataFrame(y_scaler.transform(Y.values), index=Y.index)
        
        N = H.shape[1]
        
        S_prev = model[0]
        S = model[1]
        V = model[2]
        
        omega = 1
        
        y_hat_scaled = []
        for i in range(0, len(H)):
            timepoint = H.index[i]
            
            h = H[timepoint:timepoint].values
            y = Y_scaled[timepoint:timepoint][0][0]
            
            # Predict
            if hours_ahead < 20:
                y_hat_scaled.append((h @ S)[0])
            else:
                y_hat_scaled.append((h @ S_prev)[0])
            
            # Filter
            K = (V @ t(h)) @ inv(h @ V @ t(h) + omega)
            S_prev = S
            S = S_prev + K @ (y - h @ S_prev)
            V = (identity(N) - K @ h) @ V
            
        y_hat_scaled = Series(y_hat_scaled, index=H.index)[start_date:end_date]
        y_hat = y_scaler.inverse_transform(y_hat_scaled.values.reshape(-1,1))
        y_hat = Series(y_hat.reshape(-1), index=y_hat_scaled.index)
            
        print('Hours ahead:', hours_ahead, '  ', 'MAE:',
              round(MAE(y_hat.values, Y[start_date:end_date].values)[0], 2))
        
        self.y_hat = concat([self.y_hat, y_hat])
    
    
    def _rolling_forecast_MLP(self, model):
        for day in range(0, len(self.y_test)):
            x_new = self.X_test_scaled[self.y_test.index[day]:self.y_test.index[day]]
            y_new = self.y_test_scaled[self.y_test.index[day]:self.y_test.index[day]]
            indexes = self.y_test[self.y_test.index[day]:self.y_test.index[day]].index
            
            y_hat_new = model.predict(x_new.values)
            y_hat_new = DataFrame(self.y_scaler.inverse_transform(y_hat_new),
                                  columns=self.y.columns, index=indexes)['power']
            
            self.y_hat = self.y_hat.append(y_hat_new)
            
            # Update model
            if day != 0:
                self.X_training_scaled = self.X_training_scaled.append(x_old)
                self.y_training_scaled = self.y_training_scaled.append(y_old)
                model.fit(self.X_training_scaled.values, self.y_training_scaled.values,
                          epochs=5, batch_size=256, verbose=0, shuffle=True)
                
            x_old = x_new
            y_old = y_new
        
        
    def generate_diagnostics(self):
        if self.model_type == 'MLP':
            for history in self.histories:
                history_plot(history)
        
        y_test = self.dw.get('power')['power'][self.y_hat.index[0]:self.y_hat.index[-1]]
        generate_diagnostics(y_test, self.y_hat)
        
    
    def write_to_file(self):
        y_test = self.dw.get('power')['power'][self.y_hat.index[0]:self.y_hat.index[-1]]
        y_test = y_test[datetime(2011,1,4):]
        
        write_to_file(y_test, self.y_hat, 'Direct '+self.model_type+' '+str(self.test_year),
                      self.training_time, self.update_time, True, self.validation)
        
        
    def _load_X(self, hours_ahead):
        inputs = []
        
        for hour in range(hours_ahead+self.power_lag, hours_ahead+self.power_lag+24):
            inputs.append('power shift_hours='+str(hour))
        for hour in range(hours_ahead, hours_ahead+self.power_lag+24):
            inputs.append('temp shift_hours='+str(hour))
        for hour in range(0, hours_ahead):
            inputs.append('temp_pred shift_hours='+str(hour))
            
        X = self.dw.get(inputs)
        
        # If the model is a linear regression or Kalman filter model then add
        # temperature squared and temperature cubed variables to the input
        if self.model_type == 'Lin_reg' or 'Kalman':
            for hour in range(hours_ahead, hours_ahead+self.power_lag+24):
                X['temp shift_hours='+str(hour)+' squared'] = X['temp shift_hours='+str(hour)]**2
                X['temp shift_hours='+str(hour)+' squared'] = X['temp shift_hours='+str(hour)]**3
            for hour in range(0, hours_ahead):
                X['temp_pred shift_hours='+str(hour)+' squared'] = X['temp_pred shift_hours='+str(hour)]**2
                X['temp_pred shift_hours='+str(hour)+' squared'] = X['temp_pred shift_hours='+str(hour)]**3
            
        for weekday in range(0, DAYS_PER_WEEK):
            X['day '+str(weekday)] = (X.index.weekday == weekday).astype(int)
                    
        # Only return data from a certain hour of the day
        X = X[X.index.hour == ((self.current_hour+hours_ahead) % HOURS_PER_DAY)]
        
        # Account for holidays
        start_date = X.first_valid_index().date()
        end_date = X.last_valid_index().date()
    
        holidays = generate_holidays(start_date, end_date)
        
        for holiday in holidays:
            timepoint = DatetimeIndex([holiday], tz=pytz.timezone('Europe/Oslo'), ambiguous=True)[0] + timedelta(hours=hours_ahead-14)
    
            if holiday == date(2013,3,31) or holiday == date(2016,3,27):
                if ((self.current_hour + hours_ahead) % HOURS_PER_DAY) == 2:
                    continue
                elif ((self.current_hour + hours_ahead) % HOURS_PER_DAY) != (1 or 2):
                    timepoint -= timedelta(hours=1)
                    
            X.loc[timepoint, 'day 0'] = 0
            X.loc[timepoint, 'day 1'] = 0
            X.loc[timepoint, 'day 2'] = 0
            X.loc[timepoint, 'day 3'] = 0
            X.loc[timepoint, 'day 4'] = 0
            X.loc[timepoint, 'day 5'] = 0
            X.loc[timepoint, 'day 6'] = 1
            
        if ((self.current_hour + hours_ahead) % HOURS_PER_DAY) != 1:
            X = X[1:]
            
        self.X = X
            
            
    def _load_y(self, hours_ahead):
        y = self.dw.get(['power'])[self.X.index[0]:self.X.index[-1]]
        self.y = y[y.index.hour == ((self.current_hour+hours_ahead) % HOURS_PER_DAY)]
        
        
    # Split data into training and test set  
    def _split_data(self, hours_ahead):
        test_year = self.test_year
        
        if self.validation == 'CV':
            self.X_training = self.X[self.X.index.year != test_year]
            self.y_training = self.y[self.y.index.year != test_year]
        elif self.validation == 'Expanding':
            self.X_training = self.X[self.X.index.year < test_year]
            self.y_training = self.y[self.y.index.year < test_year]
        elif self.validation == 'Sliding':
            self.X_training = self.X[self.X.index.year == (test_year-1)]
            self.y_training = self.y[self.y.index.year == (test_year-1)]
        elif 'years back' in self.validation:
            n = int(self.validation[0])
            
            self.X_training = self.X[self.X.index.year < test_year]
            self.y_training = self.y[self.y.index.year < test_year]
            
            self.X_training = self.X_training[self.X_training.index.year >= (test_year-n)]
            self.y_training = self.y_training[self.y_training.index.year >= (test_year-n)]
            
        if ((self.current_hour + hours_ahead) % 24 == 0):
            self.X_test = self.X[1:][(self.X.index.year == test_year)[:-1]]
            self.y_test = self.y[1:][(self.y.index.year == test_year)[:-1]]
        else:
            self.X_test = self.X[self.X.index.year == test_year]
            self.y_test = self.y[self.y.index.year == test_year]
            
            
    def _split_data_manually(self, training_start, training_end,
                             test_start, test_end):
        if self.validation == 'Manual':
            self.X_training = self.X[training_start:training_end]
            self.y_training = self.y[training_start:training_end]
            
            self.X_test = self.X[test_start:test_end]
            self.y_test = self.y[test_start:test_end]
            
        elif self.validation == 'CV Manual':
            self.X_training = self.X[self.X.index.year != self.test_year]
            self.y_training = self.y[self.y.index.year != self.test_year]
            
            #self.X_training = self.X_training.append(self.X[datetime(self.test_year,1,1):training_end])
            #self.y_training = self.X_training.append(self.y[datetime(self.test_year,1,1):training_end])
            
            self.X_test = self.X[test_start:test_end]
            self.y_test = self.y[test_start:test_end]
        
    
    # Standardize the training and test sets
    def _scale_data(self):
        self.X_scaler = StandardScaler()
        self.X_scaler.fit_transform(self.X_training)
        self.X_training_scaled = DataFrame(self.X_scaler.transform(self.X_training.values),
                                           index=self.X_training.index)
        self.X_test_scaled = DataFrame(self.X_scaler.transform(self.X_test.values),
                                       index=self.X_test.index)
        
        self.y_scaler = StandardScaler()
        self.y_scaler.fit_transform(self.y_training)
        self.y_training_scaled = DataFrame(self.y_scaler.transform(self.y_training.values),
                                           index=self.y_training.index)
        self.y_test_scaled = DataFrame(self.y_scaler.transform(self.y_test.values),
                                       index=self.y_test.index)