from datetime import date, datetime, timedelta
from keras.models import Model
from keras.layers import Dense, Input, LSTM, SimpleRNN, TimeDistributed
import keras.optimizers
from keras.optimizers import Adam, SGD
import numpy as np
from numpy import mean, std
from pandas import DataFrame, DatetimeIndex, Series
import pytz
from time import time

from Library.error_functions import MAE, MAPE
from Library.diagnostic_functions import generate_diagnostics, history_plot
from Library.helper_functions import generate_holidays, series_to_supervised, to_sequential, write_to_file



class Encoder_decoder_RNN_model(Model):
    def __init__(self, dw, model_type='LSTM'):
        self.dw = dw
        self.model_type = model_type
        
    
    def fit(self, test_year, validation='Expanding'):
        self.n_backward = 23
        self.n_forward = 43
        
        self.test_year = test_year
        self.validation = validation
        
        training_start_time = time()
        
        self._load_X()
        self._load_y()
        
        self.n_encoder_variables = int(self.X_encoder.shape[1]/(self.n_backward+1))
        self.n_decoder_variables = int(self.X_decoder.shape[1]/(self.n_forward+1))
        
        self._split_data()
        
        encoder_inputs = Input(shape=(self.n_backward+1, self.n_encoder_variables))
        if self.model_type == 'LSTM':
            encoder = LSTM(40, return_state=True, return_sequences=True)
        elif self.model_type == 'Elman':
            encoder = SimpleRNN(40, return_state=True, return_sequences=True)
        encoder_outputs = encoder(encoder_inputs)
        encoder_outputs, encoder_states = encoder_outputs[0], encoder_outputs[1:]
        encoder_outputs = TimeDistributed(Dense(1))(encoder_outputs)
        
        decoder_inputs = Input(shape=(self.n_forward+1, self.n_decoder_variables))
        if self.model_type == 'LSTM':
            decoder = LSTM(40, return_sequences=True)
        elif self.model_type == 'Elman':
            decoder = SimpleRNN(40, return_sequences=True)
        decoder_outputs = decoder(decoder_inputs, initial_state=encoder_states)
        decoder_outputs = TimeDistributed(Dense(1))(decoder_outputs)
        
        self.model = Model([encoder_inputs, decoder_inputs], [encoder_outputs, decoder_outputs])
        self.model.compile(optimizer=keras.optimizers.Adam(lr=0.0005), loss='mse')
        
        self.history = self.model.fit([self.X_encoder_train, self.X_decoder_train],
                            [self.y_encoder_train, self.y_decoder_train],
                            batch_size=256, epochs=1000, verbose=2,
                            validation_data=([self.X_encoder_test, self.X_decoder_test],
                                             [self.y_encoder_test, self.y_decoder_test]),
                            shuffle=True)
                             
        training_end_time = time()
        self.training_time = training_end_time - training_start_time
    
        
    def forecast(self):
        self.rolling_forecast = False
        
        update_start_time = time()
                
        self.predictions = self.model.predict([self.X_encoder_test, self.X_decoder_test])[1][:, -24:, :].reshape(-1, 24)
        self.predictions = DataFrame(self.predictions, index=self.test_indexes)
        columns = []
        for hour in range(0, 24):
            columns.append('power shift_hours='+str(hour))
        self.predictions.columns = columns
        
        self.predictions = to_sequential(self.predictions)
        self.predictions = self.std_power*self.predictions + self.mean_power
        
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
        self.update_time = update_time / len(self.y_test)
        
        return self.y_hat
    
    
    def rolling_forecast(self):
        update_start_time = time()  
            
        y_hat = Series()
        for day in range(0, len(self.test_indexes)):
            print(day)
            
            X_encoder_new = self.X_encoder_test[day].reshape(1, self.X_encoder_test.shape[1], -1)
            y_encoder_new = self.y_encoder_test[day].reshape(1, self.y_encoder_test.shape[1], -1)
            
            X_decoder_new = self.X_decoder_test[day].reshape(1, self.X_decoder_test.shape[1], -1)
            y_decoder_new = self.y_decoder_test[day].reshape(1, self.y_decoder_test.shape[1], -1)
            
            predictions = self.model.predict([X_encoder_new, X_decoder_new])[1][:, -24:, :].reshape(-1, 24)
            predictions = DataFrame(predictions, index=[self.test_indexes[day]])
            columns = []
            for hour in range(0, 24):
                columns.append('power shift_hours='+str(hour))
            predictions.columns = columns
            
            predictions = to_sequential(predictions)
            predictions = self.std_power*predictions + self.mean_power
            
            y_hat = y_hat.append(predictions)
            
            # Update model
            if day != 0:
                self.X_encoder_train = np.append(self.X_encoder_train, X_encoder_old, axis=0)
                self.y_encoder_train = np.append(self.y_encoder_train, y_encoder_old, axis=0)
                self.X_decoder_train = np.append(self.X_decoder_train, X_decoder_old, axis=0)
                self.y_decoder_train = np.append(self.y_decoder_train, y_decoder_old, axis=0)
                
                self.model.fit([self.X_encoder_train, self.X_decoder_train],
                          [self.y_encoder_train, self.y_decoder_train],
                          batch_size=256, epochs=5, verbose=0,
                          shuffle=True)
                
            X_encoder_old = X_encoder_new
            y_encoder_old = y_encoder_new
            X_decoder_old = X_decoder_new
            y_decoder_old = y_decoder_new
            
        self.y_hat = y_hat
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
        
        self.update_time = update_time / len(self.test_indexes)
        
        
    def generate_diagnostics(self):
        history_plot(self.history)
        y_test = self.dw.get('power')['power'][self.y_hat.index[0]:self.y_hat.index[-1]]
        generate_diagnostics(y_test, self.y_hat)
        
    
    def write_to_file(self):
        y_test = self.dw.get('power')['power'][self.y_hat.index[0]:self.y_hat.index[-1]]
        
        write_to_file(y_test, self.y_hat, self.model_type+' Encoder-decoder '+str(self.test_year),
                      self.training_time, self.update_time, self.rolling_forecast, self.validation)
        
        
    def _load_X(self):
        X_encoder = self.dw.get(['power shift_hours=20', 'temp shift_hours=19'])
        self.mean_power = mean(X_encoder['power shift_hours=20'])
        self.std_power = std(X_encoder['power shift_hours=20'])
        X_encoder['power shift_hours=20'] = (X_encoder['power shift_hours=20'] - self.mean_power) / self.std_power
        self.mean_temp = mean(X_encoder['temp shift_hours=19'])
        self.std_temp = std(X_encoder['temp shift_hours=19'])
        X_encoder['temp shift_hours=19'] = (X_encoder['temp shift_hours=19'] - self.mean_temp) / self.std_temp
        
        X_decoder = self.dw.get(['temp shift_hours=-'+str(23), 'temp_pred shift_hours=-'+str(23)])
        X_decoder['temp_pred shift_hours=-'+str(23)] = (X_decoder['temp_pred shift_hours=-'+str(23)] - self.mean_temp) / self.std_temp
        X_decoder['temp shift_hours=-'+str(23)] = (X_decoder['temp shift_hours=-'+str(23)] - self.mean_temp) / self.std_temp
        
        
        X_encoder = X_encoder[24:]
        X_decoder = X_decoder[24:]
        
        for X in [X_encoder, X_decoder]:
            for hour in range(0, 24):
                X['hour '+str(hour)] = (X.index.hour == hour).astype(int)
            for weekday in range(0, 7):
                X['day '+str(weekday)] = (X.index.weekday == weekday).astype(int)
                
            start_date = X.first_valid_index().date()
            end_date = X.last_valid_index().date()
        
            holidays = generate_holidays(start_date, end_date)
        
            for holiday in holidays:
                timepoint = DatetimeIndex([holiday], tz=pytz.timezone('Europe/Oslo'), ambiguous=True)[0]
                
                for hour in range(0, 24):
                    timepoint += timedelta(hours=1)
            
                    if (holiday == date(2013,3,31) and hour == 2) or (holiday == date(2016,3,27) and hour == 2):
                        pass
                    else:
                        X.loc[timepoint, 'day 0'] = 0
                        X.loc[timepoint, 'day 1'] = 0
                        X.loc[timepoint, 'day 2'] = 0
                        X.loc[timepoint, 'day 3'] = 0
                        X.loc[timepoint, 'day 4'] = 0
                        X.loc[timepoint, 'day 5'] = 0
                        X.loc[timepoint, 'day 6'] = 1
                        
        X_encoder_reframed = series_to_supervised(X_encoder, self.n_backward, 1)
        X_encoder_reframed = X_encoder_reframed[X_encoder_reframed.index.hour == 1]
        
        X_decoder_reframed = series_to_supervised(X_decoder, self.n_forward, 1)
        X_decoder_reframed = X_decoder_reframed[X_decoder_reframed.index.hour == 1]
        
        for hour in range(39, 44):
            X_decoder_reframed['var1(t-'+str(hour)+')'] = X_decoder_reframed['var2(t-'+str(hour)+')']
            
        del X_decoder_reframed['var2(t)']
        for hour in range(1, 44):
            del X_decoder_reframed['var2(t-'+str(hour)+')']
            
            
        self.X_encoder = X_encoder_reframed
        self.X_decoder = X_decoder_reframed
                
            
    def _load_y(self):
        y_encoder = self.dw.get('power shift_hours=19')
        y_encoder['power shift_hours=19'] = (y_encoder['power shift_hours=19'] - self.mean_power) / self.std_power
        
        y_decoder = self.dw.get(['power shift_hours=-'+str(23)])
        y_decoder['power shift_hours=-'+str(23)] = (y_decoder['power shift_hours=-'+str(23)] - self.mean_power) / self.std_power
        
        y_encoder_reframed = series_to_supervised(y_encoder, self.n_backward, 1)
        y_encoder_reframed = y_encoder_reframed[y_encoder_reframed.index.hour == 1]
        y_encoder_reframed = y_encoder_reframed[1:]
        
        y_decoder_reframed = series_to_supervised(y_decoder, self.n_forward, 1)
        y_decoder_reframed = y_decoder_reframed[y_decoder_reframed.index.hour == 1]
        y_decoder_reframed = y_decoder_reframed[1:]
        
        self.y_encoder = y_encoder_reframed
        self.y_decoder = y_decoder_reframed
        
        
    # Split data into training and test set  
    def _split_data(self):
        test_year = self.test_year
        
        if self.validation == 'CV':
            self.X_encoder_train = self.X_encoder[self.X_encoder.index.year != test_year]
            self.X_decoder_train = self.X_decoder[self.X_decoder.index.year != test_year]
            self.y_encoder_train = self.y_encoder[self.y_encoder.index.year != test_year]
            self.y_decoder_train = self.y_decoder[self.y_decoder.index.year != test_year]
        elif self.validation == 'Expanding':
            self.X_encoder_train = self.X_encoder[self.X_encoder.index.year < test_year]
            self.X_decoder_train = self.X_decoder[self.X_decoder.index.year < test_year]
            self.y_encoder_train = self.y_encoder[self.y_encoder.index.year < test_year]
            self.y_decoder_train = self.y_decoder[self.y_decoder.index.year < test_year]
        elif self.validation == 'Sliding':
            self.X_encoder_train = self.X_encoder[self.X_encoder.index.year == (test_year-1)]
            self.X_decoder_train = self.X_decoder[self.X_decoder.index.year == (test_year-1)]
            self.y_encoder_train = self.y_encoder[self.y_encoder.index.year == (test_year-1)]
            self.y_decoder_train = self.y_decoder[self.y_decoder.index.year == (test_year-1)]
            
        self.X_encoder_test = self.X_encoder[self.X_encoder.index.year == test_year]
        self.X_decoder_test = self.X_decoder[self.X_encoder.index.year == test_year]
        self.y_encoder_test = self.y_encoder[self.y_encoder.index.year == test_year]
        self.y_decoder_test = self.y_decoder[self.y_decoder.index.year == test_year]
        
        self.test_indexes = self.X_encoder_test.index
        
        self.X_encoder_train = self.X_encoder_train.values.reshape((self.X_encoder_train.shape[0], self.n_backward+1, self.n_encoder_variables))
        self.X_decoder_train = self.X_decoder_train.values.reshape((self.X_decoder_train.shape[0], self.n_forward+1, self.n_decoder_variables))
        self.y_encoder_train = self.y_encoder_train.values.reshape((self.y_encoder_train.shape[0], self.n_backward+1, 1))
        self.y_decoder_train = self.y_decoder_train.values.reshape((self.y_decoder_train.shape[0], self.n_forward+1, 1))
    
        self.X_encoder_test = self.X_encoder_test.values.reshape((self.X_encoder_test.shape[0], self.n_backward+1, self.n_encoder_variables))
        self.X_decoder_test = self.X_decoder_test.values.reshape((self.X_decoder_test.shape[0], self.n_forward+1, self.n_decoder_variables))
        self.y_encoder_test = self.y_encoder_test.values.reshape((self.y_encoder_test.shape[0], self.n_backward+1, 1))
        self.y_decoder_test = self.y_decoder_test.values.reshape((self.y_decoder_test.shape[0], self.n_forward+1, 1))
        
            
    
    # Standardize the training and test sets
    def _scale_data(self):
        # done elsewhere
        pass