from datetime import timedelta
from keras.layers import Dense
from keras.models import Sequential
from keras.optimizers import Adam, SGD
from pandas import concat, DataFrame, DatetimeIndex, Series
import pytz
from sklearn.preprocessing import StandardScaler
from time import time

from Library.diagnostic_functions import generate_diagnostics, history_plot
from Library.helper_functions import generate_holidays, to_sequential, write_to_file

from Models.model import Model


class MIMO_MLP_model(Model):
    def __init__(self, dw):
        self.dw = dw
        
    
    def fit(self, test_year, validation='Expanding',
            training_start=None, training_end=None, test_start=None, test_end=None):
        self.training_start = training_start
        self.training_end = training_end
        self.test_start = test_start
        self.test_end = test_end
        
        self.test_year = test_year
        self.validation = validation
        
        training_start_time = time()
        
        self._load_X()
        self._load_y()
        
        if self.validation == 'Manual':
            self._split_data_manually(self.training_start, self.training_end, 
                                      self.test_start, self.test_end)
        else:
            self._split_data()
            
        self._scale_data()
        
        
        self.model = Sequential()
        self.model.add(Dense(40, input_dim=self.X.shape[1],
                                      kernel_initializer='normal', activation='relu'))
        self.model.add(Dense(24, kernel_initializer='normal'))
        self.model.compile(loss='mse', optimizer=Adam(lr=0.0001))
        
        self.history = self.model.fit(self.X_training_scaled.values,
                                      self.y_training_scaled.values,
                             validation_data=(self.X_test_scaled.values,
                                              self.y_test_scaled.values),
                             epochs=2000, batch_size=256, verbose=2, shuffle=True)
                             
        training_end_time = time()
        self.training_time = training_end_time - training_start_time
    
        
    def forecast(self):
        self.rolling_forecast = False
        
        update_start_time = time()  
        
        self.y_hat = self.model.predict(self.X_test_scaled.values)
        self.y_hat = DataFrame(self.y_scaler.inverse_transform(self.y_hat),
                                columns=self.y.columns, index=self.y_test.index)
        self.y_hat = to_sequential(self.y_hat)
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
        self.update_time = update_time / len(self.y_test)
        
        return self.y_hat
    
    
    def rolling_forecast(self):
        self.rolling_forecast = True
        
        update_start_time = time()
        
        y_hat = Series()
        for day in range(0, len(self.y_test)):
            print(day)
            
            x_new = self.X_test_scaled[self.y_test.index[day]:self.y_test.index[day]]
            y_new = self.y_test_scaled[self.y_test.index[day]:self.y_test.index[day]]
            indexes = y_new.index
            
            y_hat_new = self.model.predict(x_new.values)
            y_hat_new = DataFrame(self.y_scaler.inverse_transform(y_hat_new),
                                    columns=self.y.columns, index=indexes)
            y_hat_new = to_sequential(y_hat_new)
            
            y_hat = y_hat.append(y_hat_new)
            
            # Update model
            if day != 0:
                self.X_training_scaled = self.X_training_scaled.append(x_old)
                self.y_training_scaled = self.y_training_scaled.append(y_old)
                self.model.fit(self.X_training_scaled.values,
                               self.y_training_scaled.values,
                               epochs=5, batch_size=256, verbose=0, shuffle=True)
                
            x_old = x_new
            y_old = y_new
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
        self.update_time = update_time / len(self.y_test)
        
        self.y_hat = y_hat
        
        return self.y_hat
        
        
    def generate_diagnostics(self):
        history_plot(self.history, ymax=0.04)
        generate_diagnostics(to_sequential(self.y_test), self.y_hat)
        
    
    def write_to_file(self):
        write_to_file(to_sequential(self.y_test), self.y_hat, 'Alt MIMO MLP '+str(self.test_year),
                      self.training_time, self.update_time, self.rolling_forecast, self.validation)
        
        
    def _load_X(self):
        K = 5
        L = 24
        M = 15
        
        inputs = []
        
        for hour in range(M+K, M+K+L):
            inputs.append('power shift_hours='+str(hour))
        for hour in range(M, M+K+L):
            inputs.append('temp shift_hours='+str(hour))
        for hour in range(-M+1, L):
            inputs.append('temp_pred shift_hours='+str(-hour))
            
        X = self.dw.get(inputs)
    
        for weekday in range(0, 7):
            X['day '+str(weekday)] = (X.index.weekday == weekday).astype(int)
        
        X = X[X.index.hour == 1]
        
        # Account for public holidays
        start_date = X.first_valid_index().date()
        end_date = X.last_valid_index().date()
        
        holidays = generate_holidays(start_date, end_date)
        for holiday in holidays:
            timepoint = DatetimeIndex([holiday], tz=pytz.timezone('Europe/Oslo'))[0] + timedelta(hours=1)
            
            X.loc[timepoint, 'day 0'] = 0
            X.loc[timepoint, 'day 1'] = 0
            X.loc[timepoint, 'day 2'] = 0
            X.loc[timepoint, 'day 3'] = 0
            X.loc[timepoint, 'day 4'] = 0
            X.loc[timepoint, 'day 5'] = 0
            X.loc[timepoint, 'day 6'] = 1
        
        self.X = X
            
            
    def _load_y(self):
        y1 = self.dw.get('power shift_hours=0')
              
        inputs = []
        for hour in range(1, 10):
            inputs.append('power shift_hours='+str(-hour))
        y2 = self.dw.get(inputs)    
        y = concat([y1, y2], axis=1)
        
        inputs = []
        for hour in range(10, 24):
            inputs.append('power shift_hours='+str(-hour))
        y3 = self.dw.get(inputs)        
        y = concat([y, y3], axis=1)
                
        y = y[y.index.hour == 1]
        
        y = y[2:]
        
        self.y = y
        
        
    # Split data into training and test set  
    def _split_data(self):
        test_year = self.test_year
        
        if self.validation == 'CV':
            self.X_training = self.X[self.X.index.year != test_year]
            self.y_training = self.y[self.y.index.year != test_year]
        elif self.validation == 'Expanding':
            self.X_training = self.X[self.X.index.year < test_year]
            self.y_training = self.y[self.y.index.year < test_year]
        elif self.validation == 'Sliding':
            self.X_training = self.X[self.X.index.year == (test_year-1)]
            self.y_training = self.y[self.y.index.year == (test_year-1)]
        elif 'years back' in self.validation:
            n = int(self.validation[0])
            
            self.X_training = self.X[(self.X.index.year < test_year)]
            self.y_training = self.y[(self.y.index.year < test_year)]
            
            self.X_training = self.X_training[self.X_training.index.year >= (test_year-n)]
            self.y_training = self.y_training[self.y_training.index.year >= (test_year-n)]
            
        self.X_test = self.X[self.X.index.year == test_year]
        self.y_test = self.y[self.y.index.year == test_year]
        
        
    def _split_data_manually(self, training_start, training_end,
                             test_start, test_end):
        self.X_training = self.X[training_start:training_end]
        self.y_training = self.y[training_start:training_end]
        
        self.X_test = self.X[test_start:test_end]
        self.y_test = self.y[test_start:test_end]
        
    
    # Standardize the training and test sets
    def _scale_data(self):
        self.X_scaler = StandardScaler()
        self.X_scaler.fit_transform(self.X_training)
        self.X_training_scaled = DataFrame(self.X_scaler.transform(self.X_training.values),
                                           index=self.X_training.index)
        self.X_test_scaled = DataFrame(self.X_scaler.transform(self.X_test.values),
                                       index=self.X_test.index)
        
        self.y_scaler = StandardScaler()
        self.y_scaler.fit_transform(self.y_training)
        self.y_training_scaled = DataFrame(self.y_scaler.transform(self.y_training.values),
                                           index=self.y_training.index)
        self.y_test_scaled = DataFrame(self.y_scaler.transform(self.y_test.values),
                                       index=self.y_test.index)