from datetime import date, datetime, timedelta
from keras.models import Sequential
from keras.layers import Dense
from keras.optimizers import Adam, SGD
from pandas import concat, DataFrame, date_range, DatetimeIndex, Series
import pytz
from sklearn.preprocessing import StandardScaler
from time import time

from Library.error_functions import MAE, MAPE
from Library.diagnostic_functions import generate_diagnostics, history_plot
from Library.helper_functions import generate_holidays, to_sequential, write_to_file

from Models.model import Model


HOURS_PER_DAY = 24
DAYS_PER_WEEK = 7
HOURS_PER_WEEK = HOURS_PER_DAY * DAYS_PER_WEEK


class SMSO_MLP_model(Model):
    def __init__(self, dw):
        self.dw = dw
        
    
    def fit(self, test_year, validation='Expanding'):
        self.test_year = test_year
        self.validation = validation
        
        training_start_time = time()
        
        self._load_X()
        self._load_y()
        self._split_data()
        self._scale_data()
        
        
        self.model = Sequential()
        self.model.add(Dense(40, input_dim=self.X.shape[1],
                             kernel_initializer='normal', activation='relu'))
        self.model.add(Dense(1))
        self.model.compile(loss='mean_squared_error', optimizer=Adam(lr=0.0001))
        self.history = self.model.fit(self.X_training_scaled.values, self.y_training_scaled.values,
                            validation_data=(self.X_test_scaled.values, self.y_test_scaled.values),
                            epochs=200, batch_size=256, verbose=2)
                             
        training_end_time = time()
        self.training_time = training_end_time - training_start_time
    
        
    def forecast(self):
        self.rolling_forecast = False
        
        update_start_time = time()  
        
        self.y_hat = self.model.predict(self.X_test_scaled.values)[0]
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
        self.update_time = update_time / len(self.y_test)
        
        return self.y_hat
    
    
    def rolling_forecast(self):
        self.rolling_forecast = True
        
        update_start_time = time()  
        
        y_hat = Series()
        
        days_in_year = len(date_range(self.y_test.index[0], self.y_test.index[-1], freq='d'))
        for day in range(0, days_in_year):
            print(day)
            
            indexes = date_range(datetime(self.test_year,1,1,1) + timedelta(day),
                                 datetime(self.test_year,1,2,0) + timedelta(day),
                                 freq='h', tz='Europe/Oslo')
            
            x_new = self.X_test_scaled[indexes[0]:indexes[-1]]
            y_new = self.y_test_scaled[indexes[0]:indexes[-1]]
            
            y_hat_new = self.model.predict(x_new.values)
            y_hat_new = DataFrame(self.y_scaler.inverse_transform(y_hat_new), index=indexes)
            
            y_hat = y_hat.append(y_hat_new)
            
            # Update model
            if day != 0:
                self.X_training_scaled = self.X_training_scaled.append(x_old)
                self.y_training_scaled = self.y_training_scaled.append(y_old)
                
                self.model.fit(self.X_training_scaled.values,
                               self.y_training_scaled.values, epochs=1,
                               batch_size=256, verbose=0)
                
            x_old = x_new
            y_old = y_new
            
        update_end_time = time()
        update_time = update_end_time - update_start_time
    
        self.update_time = update_time / days_in_year
        
        self.y_hat = y_hat[0]
        
        return self.y_hat
        
        
    def generate_diagnostics(self):
        history_plot(self.history, ymax=0.04)
        generate_diagnostics(to_sequential(self.y_test), self.y_hat)
        
    
    def write_to_file(self):
        write_to_file(self.y_test['power'], self.y_hat, 'SMSO MLP '+str(self.test_year),
                      self.training_time, self.update_time, self.rolling_forecast, self.validation)
        
        
    def _load_X(self):
        current_hour=5
        
        max_hours_ahead = (HOURS_PER_DAY - current_hour) + HOURS_PER_DAY
    
        
        prediction_hour=24
        hours_ahead = HOURS_PER_DAY - current_hour + HOURS_PER_DAY
        
        inputs = []
        for hour in range(hours_ahead, hours_ahead+HOURS_PER_DAY):
            inputs.append('power shift_hours='+str(hour))
        for hour in range(hours_ahead, hours_ahead+HOURS_PER_DAY):
            inputs.append('temp shift_hours='+str(hour))
        for hour in range(-HOURS_PER_DAY+prediction_hour, prediction_hour+(HOURS_PER_DAY-current_hour)):
            inputs.append('temp_pred shift_hours='+str(hour))
            
        X = self.dw.get(inputs)
        X = X[X.index.hour == 0]
        
        for n in range(0, HOURS_PER_DAY):
            X['power '+str(n)] = X['power shift_hours='+str(hours_ahead+n)]
            X['temp '+str(n)] = X['temp shift_hours='+str(hours_ahead+n)]
            del X['power shift_hours='+str(hours_ahead+n)]
            del X['temp shift_hours='+str(hours_ahead+n)]
        for n in range(0, max_hours_ahead):
            X['temp_pred '+str(n)] = X['temp_pred shift_hours='+str(n-HOURS_PER_DAY+prediction_hour)]
            del X['temp_pred shift_hours='+str(n-HOURS_PER_DAY+prediction_hour)]
        
        
        for prediction_hour in range(1, HOURS_PER_DAY):
            prediction_hour
            hours_ahead = HOURS_PER_DAY - current_hour + prediction_hour
            
            inputs = []
            for hour in range(hours_ahead, hours_ahead+HOURS_PER_DAY):
                inputs.append('power shift_hours='+str(hour))
                inputs.append('temp shift_hours='+str(hour))
            for hour in range(-HOURS_PER_DAY+prediction_hour, prediction_hour+(HOURS_PER_DAY-current_hour)):
                inputs.append('temp_pred shift_hours='+str(hour))
                
            X1 = self.dw.get(inputs)
            X1 = X1[X1.index.hour == prediction_hour]
            
            for n in range(0, HOURS_PER_DAY):
                X1['power '+str(n)] = X1['power shift_hours='+str(hours_ahead+n)]
                X1['temp '+str(n)] = X1['temp shift_hours='+str(hours_ahead+n)]
                del X1['power shift_hours='+str(hours_ahead+n)]
                del X1['temp shift_hours='+str(hours_ahead+n)]
            for n in range(0, max_hours_ahead):
                X1['temp_pred '+str(n)] = X1['temp_pred shift_hours='+str(n-HOURS_PER_DAY+prediction_hour)]
                del X1['temp_pred shift_hours='+str(n-HOURS_PER_DAY+prediction_hour)]
                
            X = concat([X, X1], axis=0)
            
        X = X.sort_index()
            
        for weekday in range(0, DAYS_PER_WEEK):
            X['day '+str(weekday)] = (X.index.weekday == weekday).astype(int)
            
        for hour in range(0, HOURS_PER_DAY):
            X['hour '+str(hour)] = (X.index.hour == hour).astype(int)
            
        start_date = X.first_valid_index().date()
        end_date = X.last_valid_index().date()
    
        holidays = generate_holidays(start_date, end_date)
    
        for holiday in holidays:
            timepoint = DatetimeIndex([holiday], tz=pytz.timezone('Europe/Oslo'), ambiguous=True)[0]
            
            for hour in range(0, 24):
                timepoint += timedelta(hours=1)
        
                if holiday == date(2013,3,31) or holiday == date(2016,3,27):
                    if hour == 2:
                        continue
                    elif hour > 2:
                        timepoint -= timedelta(hours=1)
                        
                X.loc[timepoint, 'day 0'] = 0
                X.loc[timepoint, 'day 1'] = 0
                X.loc[timepoint, 'day 2'] = 0
                X.loc[timepoint, 'day 3'] = 0
                X.loc[timepoint, 'day 4'] = 0
                X.loc[timepoint, 'day 5'] = 0
                X.loc[timepoint, 'day 6'] = 1
                
        self.X = X[HOURS_PER_DAY*5:]
                
            
    def _load_y(self):
        self.y = self.dw.get('power')[self.X.index[0]:self.X.index[-1]]
        
        
    # Split data into training and test set  
    def _split_data(self):
        test_year = self.test_year
        
        if self.validation == 'CV':
            self.X_training = self.X[1:][(self.X.index.year != test_year)[:-1]]
            self.y_training = self.y[1:][(self.y.index.year != test_year)[:-1]]
        elif self.validation == 'Expanding':
            self.X_training = self.X[1:][(self.X.index.year < test_year)[:-1]]
            self.y_training = self.y[1:][(self.y.index.year < test_year)[:-1]]
        elif self.validation == 'Sliding':
            self.X_training = self.X[1:][(self.X.index.year == (test_year-1))[:-1]]
            self.y_training = self.y[1:][(self.y.index.year == (test_year-1))[:-1]]
        elif 'years back' in self.validation:
            n = int(self.validation[0])
            
            self.X_training = self.X[1:][(self.X.index.year < test_year)[:-1]]
            self.y_training = self.y[1:][(self.y.index.year < test_year)[:-1]]
            
            self.X_training = self.X_training[self.X_training.index.year >= (test_year-n)]
            self.y_training = self.y_training[self.y_training.index.year >= (test_year-n)]

        self.X_test = self.X[1:][(self.X.index.year == test_year)[:-1]]
        self.y_test = self.y[1:][(self.y.index.year == test_year)[:-1]]
        
    
    # Standardize the training and test sets
    def _scale_data(self):
        self.X_scaler = StandardScaler()
        self.X_scaler.fit_transform(self.X_training)
        self.X_training_scaled = DataFrame(self.X_scaler.transform(self.X_training.values),
                                           index=self.X_training.index)
        self.X_test_scaled = DataFrame(self.X_scaler.transform(self.X_test.values),
                                       index=self.X_test.index)
        
        self.y_scaler = StandardScaler()
        self.y_scaler.fit_transform(self.y_training)
        self.y_training_scaled = DataFrame(self.y_scaler.transform(self.y_training.values),
                                           index=self.y_training.index)
        self.y_test_scaled = DataFrame(self.y_scaler.transform(self.y_test.values),
                                       index=self.y_test.index)