from datetime import datetime, timedelta
from pandas import Series
from seaborn import set_style

from Library.datawrapper import DataWrapper
from Library.diagnostic_functions import errors_by_hour_plot, generate_diagnostics
from Library.diagnostic_functions import plot_comparison
from Library.helper_functions import load_all_rolling_predictions_for_year, write_to_file

from Models.direct_model import Direct_model
from Models.encoder_decoder_RNN_model import Encoder_decoder_RNN_model
from Models.mimo_mlp_model import MIMO_MLP_model
from Models.smso_mlp_model import SMSO_MLP_model

set_style("darkgrid")



def main():
    dw = DataWrapper()
    
    validation_strategy = 'Expanding'
    
    for test_year in range(2012, 2018):
        for model in [MIMO_MLP_model(dw),
                      SMSO_MLP_model(dw),
                      Direct_model(dw, model_type='MLP'),
                      Direct_model(dw, model_type='Kalman'),
                      Encoder_decoder_RNN_model(dw, model_type='Elman'),
                      Encoder_decoder_RNN_model(dw, model_type='LSTM')]:
            model.fit(test_year, validation_strategy)
            model.rolling_forecast()
            model.write_to_file()
            
        for model_type in ['Lin_reg', 'SVR']:
            continously_retrain_model(dw, test_year, model_type, days_between_retraining=7)
            
                
    y_hat = load_all_rolling_predictions_for_year(dw, 2017)['Ensemble']
    y_test = dw.get('power')[y_hat.index[0]:y_hat.index[-1]]['power']

    generate_diagnostics(y_test, y_hat)
    errors_by_hour_plot(y_test, y_hat)

    plot_comparison(y_test, y_hat, start_date=datetime(2017, 11, 13, 1),
                    end_date=datetime(2017, 12, 4, 0))
    
    



# Function for testing a model by countinously retraining it as new
# data becomes available
def continously_retrain_model(dw, test_year, model_type, days_between_retraining=7,
                              cross_validation=False):
    training_time = 0
    
    y_hat = Series()
    
    training_start = datetime(2011, 1, 1)
    training_end = datetime(test_year, 1, 1, 0)
    test_start = datetime(test_year, 1, 1, 1)
    test_end = datetime(test_year, 1, 1, 0) + timedelta(days_between_retraining)
    
    while True:
        if test_end.year != test_year:
            test_end = datetime(test_year+1, 1, 1, 0)
        
        print('\nTest start:', test_start, '  Test end:', test_end, '\n')
        
        if model_type == 'MLP':
            model = MIMO_MLP_model(dw)
        else:
            model = Direct_model(dw, model_type)
        
        if cross_validation:
            test_approach = 'CV Manual'
        else:
            test_approach = 'Manual'
        
        model.fit(test_year, test_approach, training_start, training_end,
                  test_start, test_end)
        y_hat_new = model.forecast()
        y_hat = y_hat.append(y_hat_new)
        
        training_time += model.training_time
        
        if test_end.year != test_year:
            break
        
        training_end += timedelta(days_between_retraining)
        test_start += timedelta(days_between_retraining)
        test_end += timedelta(days_between_retraining)
        
    y_hat = y_hat[datetime(2011,1,4):]
    
    y_test = dw.get('power')['power'][y_hat.index[0]:y_hat.index[-1]]
    generate_diagnostics(y_test, y_hat)
    
    write_to_file(y_test, y_hat, 'Direct '+model_type+' '+str(test_year), training_time, -1, False, 'Expanding retraining')
    
    
    
main()